@@ -14,7 +14,10 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 import click
-from snapcraft.internal import remote_parts
+
+from snapcraft.internal import remote_parts, lifecycle
+from ._options import get_project_options
+from . import env
 
 
 @click.group(context_settings={})
@@ -25,8 +28,16 @@ def partscli(ctx):
 
 @partscli.command()
 @click.pass_context
-def update(ctx):
+def update(ctx, **kwargs):
     """Updates the parts listing from the cloud."""
+    # Update in the container so that it will use the parts at build time
+    build_environment = env.BuilderEnvironmentConfig()
+    if not build_environment.is_host:
+        project_options = get_project_options(**kwargs)
+        lifecycle.containerbuild('update', project_options)
+
+    # Parts can be defined and searched from any folder on the host, so
+    # regardless of using containers we always update these as well
     remote_parts.update()
 
 
