@@ -15,8 +15,6 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 # Data/methods shared between plugins and snapcraft
-
-from contextlib import suppress
 import glob
 import logging
 import math
@@ -26,6 +24,8 @@ import subprocess
 import sys
 import tempfile
 import urllib
+from contextlib import suppress
+from typing import List  # noqa
 
 from snapcraft.internal import errors
 
@@ -40,10 +40,11 @@ _schemadir = _DEFAULT_SCHEMADIR
 _DEFAULT_LIBRARIESDIR = os.path.join(sys.prefix, 'share', 'snapcraft',
                                      'libraries')
 _librariesdir = _DEFAULT_LIBRARIESDIR
+_DOCKERENV_FILE = '/.dockerenv'
 
 MAX_CHARACTERS_WRAP = 120
 
-env = []
+env = []  # type: List[str]
 
 logger = logging.getLogger(__name__)
 
@@ -80,12 +81,23 @@ def run_output(cmd, **kwargs):
             return output.decode('latin-1', 'surrogateescape').strip()
 
 
-def get_core_path():
-    """Returns the path to the core snap."""
-    return os.path.join(os.path.sep, 'snap', 'core', 'current')
+def get_core_path(base):
+    """Returns the path to the core base snap."""
+    return os.path.join(os.path.sep, 'snap', base, 'current')
 
 
-def format_snap_name(snap):
+def format_snap_name(snap, *, allow_empty_version: bool=False) -> str:
+    """Return a filename representing the snap depending on snap attributes.
+
+    :param dict snap: a dictionary of keys containing name, version and arch.
+    :param bool allow_empty_version: if set a filename without a version is
+                                     allowed.
+    """
+    if allow_empty_version and snap.get('version') is None:
+        template = '{name}_{arch}.snap'
+    else:
+        template = '{name}_{version}_{arch}.snap'
+
     if 'arch' not in snap:
         snap['arch'] = snap.get('architectures', None)
     if not snap['arch']:
@@ -95,11 +107,19 @@ def format_snap_name(snap):
     else:
         snap['arch'] = 'multi'
 
-    return '{name}_{version}_{arch}.snap'.format(**snap)
+    return template.format(**snap)
 
 
-def is_snap():
-    return os.environ.get('SNAP_NAME') == 'snapcraft'
+def is_snap() -> bool:
+    snap_name = os.environ.get('SNAP_NAME', '')
+    is_snap = snap_name == 'snapcraft'
+    logger.debug('snapcraft is running as a snap {!r}, '
+                 'SNAP_NAME set to {!r}'.format(is_snap, snap_name))
+    return is_snap
+
+
+def is_docker_instance() -> bool:
+    return os.path.exists(_DOCKERENV_FILE)
 
 
 def set_plugindir(plugindir):
