@@ -25,27 +25,40 @@ import requests_unixsocket
 import shutil
 import sys
 import tempfile
-import yaml
+import contextlib
 from contextlib import contextmanager
 import subprocess
 import time
 from urllib import parse
+from textwrap import dedent
+from typing import List
 
 from snapcraft.internal import common
 from snapcraft.internal.errors import (
         ContainerConnectionError,
+        ContainerError,
         ContainerRunError,
         ContainerSnapcraftCmdError,
+        InvalidContainerImageInfoError,
         SnapdError,
 )
-from snapcraft._options import _get_deb_arch
+from snapcraft.project._project_options import _get_deb_arch
 from snapcraft.internal.repo import snaps
 
 logger = logging.getLogger(__name__)
 
-_NETWORK_PROBE_COMMAND = \
-    'import urllib.request; urllib.request.urlopen("{}", timeout=5)'.format(
-        'http://start.ubuntu.com/connectivity-check.html')
+_NETWORK_PROBE_COMMAND = dedent('''
+    import urllib.request
+    import sys
+
+    check_url = "http://start.ubuntu.com/connectivity-check.html"
+    try:
+        urllib.request.urlopen(check_url, timeout=5)
+    except urllib.error.URLError as e:
+        sys.exit('Failed to open {!r}: {!s}'.format(check_url, e.reason))
+    except Exception as e:
+        sys.exit('Failed to open {!r}: {!s}'.format(check_url, e))
+    ''')
 _PROXY_KEYS = ['http_proxy', 'https_proxy', 'no_proxy', 'ftp_proxy']
 # Canonical store account key
 _STORE_KEY = (
@@ -54,41 +67,25 @@ _STORE_KEY = (
 
 class Containerbuild:
 
-    def __init__(self, *, output, source, project_options,
-                 metadata, container_name, remote=None):
-        if not output:
-            output = common.format_snap_name(metadata)
+    def __init__(self, *, source, project_options, metadata,
+                 container_name, output=None, remote=None):
+        if output is None:
+            output = common.format_snap_name(
+                metadata, allow_empty_version=True)
         self._snap_output = output
         self._source = os.path.realpath(source)
         self._project_options = project_options
         self._metadata = metadata
+        self._user = 'root'
         self._project_folder = '/root/build_{}'.format(metadata['name'])
 
-        if not remote:
-            remote = _get_default_remote()
-        _verify_remote(remote)
-        self._container_name = '{}:snapcraft-{}'.format(remote, container_name)
-        server_environment = self._get_remote_info()['environment']
-        # Use the server architecture to avoid emulation overhead
-        try:
-            kernel = server_environment['kernel_architecture']
-        except KeyError:
-            kernel = server_environment['kernelarchitecture']
-        self._server_arch = _get_deb_arch(kernel)
-        if not self._server_arch:
-            raise ContainerConnectionError(
-                'Unrecognized server architecture {}'.format(kernel))
-        self._image = 'ubuntu:xenial/{}'.format(self._server_arch)
+        self._remote = remote
+        self._container_name = 'snapcraft-{}'.format(container_name)
+        self._image = 'ubuntu:xenial'
         # Use a temporary folder the 'lxd' snap can access
-        lxd_common_dir = os.path.expanduser(
+        self._lxd_common_dir = os.path.expanduser(
             os.path.join('~', 'snap', 'lxd', 'common'))
-        os.makedirs(lxd_common_dir, exist_ok=True)
-        self.tmp_dir = tempfile.mkdtemp(prefix='snapcraft', dir=lxd_common_dir)
-
-    def _get_remote_info(self):
-        remote = self._container_name.split(':')[0]
-        return yaml.load(subprocess.check_output([
-            'lxc', 'info', '{}:'.format(remote)]).decode())
+        os.makedirs(self._lxd_common_dir, exist_ok=True)
 
     @contextmanager
     def _container_running(self):
@@ -102,12 +99,18 @@ class Containerbuild:
                 else:
                     raise e
             else:
-                # Remove temporary folder if everything went well
-                shutil.rmtree(self.tmp_dir)
                 self._finish()
 
+    def _ensure_remote(self):
+        if not self._remote:
+            self._remote = _get_default_remote()
+        _verify_remote(self._remote)
+        self._container_name = '{}:{}'.format(self._remote,
+                                              self._container_name)
+
     @contextmanager
     def _ensure_started(self):
+        self._ensure_remote()
         try:
             self._ensure_container()
             yield
@@ -148,8 +151,12 @@ class Containerbuild:
             'Failed to get container image info: {}\n'
             'It will not be recorded in manifest.')
         try:
+            # This command takes the same image name as used to create a new
+            # container. But we must always use the form distro:series/arch
+            # here so that we get only the image we're actually using!
             image_info_command = [
-                'lxc', 'image', 'list', '--format=json', self._image]
+                'lxc', 'image', 'list', '--format=json',
+                '{}/{}'.format(self._image, self._get_container_arch())]
             image_info = json.loads(subprocess.check_output(
                 image_info_command).decode())
         except subprocess.CalledProcessError as e:
@@ -167,6 +174,15 @@ class Containerbuild:
         for field in ('fingerprint', 'architecture', 'created_at'):
             if field in image_info[0]:
                 edited_image_info[field] = image_info[0][field]
+
+        # Pick up existing image info if set
+        image_info_str = os.environ.get('SNAPCRAFT_IMAGE_INFO')
+        if image_info_str:
+            try:
+                edited_image_info.update(json.loads(image_info_str))
+            except json.decoder.JSONDecodeError as e:
+                raise InvalidContainerImageInfoError(image_info_str) from e
+
         # Pass the image info to the container so it can be used when recording
         # information about the build environment.
         subprocess.check_call([
@@ -186,9 +202,10 @@ class Containerbuild:
                 command += ['--target-arch', self._project_options.target_arch]
             if args:
                 command += args
-            self._container_run(command, cwd=self._project_folder)
+            self._container_run(command, cwd=self._project_folder,
+                                user=self._user)
 
-    def _container_run(self, cmd, cwd=None, **kwargs):
+    def _container_run(self, cmd: List[str], cwd=None, user='root', **kwargs):
         sh = ''
         original_cmd = cmd.copy()
         # Automatically wait on lock files before running commands
@@ -203,6 +220,8 @@ class Containerbuild:
         if sh:
             cmd = ['sh', '-c', '{}{}'.format(sh,
                    ' '.join(pipes.quote(arg) for arg in cmd))]
+        if user != 'root':
+            cmd = ['sudo', '-H', '-E', '-u', user] + cmd
         try:
             subprocess.check_call([
                 'lxc', 'exec', self._container_name, '--'] + cmd,
@@ -227,21 +246,26 @@ class Containerbuild:
             except ContainerRunError as e:
                 retry_count -= 1
                 if retry_count == 0:
-                    raise e
+                    raise ContainerConnectionError(
+                        'No network connection in the container.\n'
+                        'If using a proxy, check its configuration.')
         logger.info('Network connection established')
 
-    def _inject_snapcraft(self):
+    def _inject_snapcraft(self, *, new_container: bool):
         if common.is_snap():
-            # Because of https://bugs.launchpad.net/snappy/+bug/1628289
-            self._container_run(['apt-get', 'install', 'squashfuse', '-y'])
-
-            # Push core snap into container
-            self._inject_snap('core')
-            self._inject_snap('snapcraft')
-        else:
+            with tempfile.TemporaryDirectory(
+                    prefix='snapcraft', dir=self._lxd_common_dir) as tmp_dir:
+                # Wait for any on-going refreshes to finish.
+                # If there are no changes an error will be returned.
+                with contextlib.suppress(ContainerRunError):
+                    self._container_run([
+                        'snap', 'watch', '--last=auto-refresh'])
+                self._inject_snap('core', tmp_dir)
+                self._inject_snap('snapcraft', tmp_dir)
+        elif new_container:
             self._container_run(['apt-get', 'install', 'snapcraft', '-y'])
 
-    def _inject_snap(self, name):
+    def _inject_snap(self, name: str, tmp_dir: str):
         session = requests_unixsocket.Session()
         # Cf. https://github.com/snapcore/snapd/wiki/REST-API#get-v2snapsname
         # TODO use get_local_snap info from the snaps module.
@@ -261,8 +285,8 @@ class Containerbuild:
         is_classic = json['result']['confinement'] == 'classic'
 
         # If the server has a different arch we can't inject local snaps
-        if (self._project_options.target_arch
-                and self._project_options.target_arch != self._server_arch):
+        target_arch = self._project_options.target_arch
+        if (target_arch and target_arch != self._get_container_arch()):
             channel = json['result']['channel']
             return self._install_snap(name, channel, is_classic=is_classic)
 
@@ -277,7 +301,7 @@ class Containerbuild:
         installed = os.path.join(os.path.sep, 'var', 'lib', 'snapd', 'snaps',
                                  filename)
 
-        filepath = os.path.join(self.tmp_dir, filename)
+        filepath = os.path.join(tmp_dir, filename)
         if rev.startswith('x'):
             logger.info('Making {} user-accessible'.format(filename))
             subprocess.check_call(['sudo', 'cp', installed, filepath])
@@ -296,7 +320,7 @@ class Containerbuild:
                 ['snap-declaration', 'snap-name={}'.format(name)],
                 ['snap-revision', 'snap-revision={}'.format(rev),
                  'snap-id={}'.format(id)],
-            ])
+            ], tmp_dir)
 
         container_filename = os.path.join(os.sep, 'run', filename)
         self._push_file(filepath, container_filename)
@@ -304,6 +328,14 @@ class Containerbuild:
                            is_dangerous=rev.startswith('x'),
                            is_classic=is_classic)
 
+    def _get_container_arch(self):
+        info = subprocess.check_output([
+            'lxc', 'info', self._container_name]).decode('utf-8')
+        for line in info.splitlines():
+            if line.startswith("Architecture:"):
+                return _get_deb_arch(line.split(None, 1)[1].strip())
+        raise ContainerError("Could not find architecture for container")
+
     def _pull_file(self, src, dst):
         subprocess.check_call(['lxc', 'file', 'pull',
                                '{}{}'.format(self._container_name, src), dst])
@@ -352,8 +384,9 @@ class Containerbuild:
             checksum_container = None
         return checksum == checksum_container
 
-    def _inject_assertions(self, filename, assertions):
-        filepath = os.path.join(self.tmp_dir, filename)
+    def _inject_assertions(self, filename: str,
+                           assertions: List[List[str]], tmp_dir: str):
+        filepath = os.path.join(tmp_dir, filename)
         with open(filepath, 'wb') as f:
             for assertion in assertions:
                 logger.info('Looking up assertion {}'.format(assertion))
