@@ -1,6 +1,6 @@
 # -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
 #
-# Copyright (C) 2015-2017 Canonical Ltd
+# Copyright (C) 2015-2018 Canonical Ltd
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License version 3 as
@@ -14,6 +14,7 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+import collections
 import contextlib
 import copy
 import filecmp
@@ -22,26 +23,20 @@ import os
 import shutil
 import sys
 from glob import glob, iglob
+from typing import Dict, Set, Sequence  # noqa: F401
 
 import yaml
 
-import snapcraft
+import snapcraft.extractors
 from snapcraft import file_utils
-from snapcraft.internal import (
-    common,
-    errors,
-    libraries,
-    repo,
-    sources,
-    states,
-)
-from ._scriptlets import ScriptRunner
+from snapcraft.internal import common, elf, errors, repo, sources, states
+from snapcraft.internal.mangling import clear_execstack
+
 from ._build_attributes import BuildAttributes
-
-if sys.platform == 'linux':
-    import magic
-
+from ._metadata_extraction import extract_metadata
 from ._plugin_loader import load_plugin  # noqa
+from ._runner import Runner
+from ._patchelf import PartPatcher
 
 logger = logging.getLogger(__name__)
 
@@ -64,15 +59,28 @@ class PluginHandler:
 
     def __init__(self, *, plugin, part_properties, project_options,
                  part_schema, definitions_schema, stage_packages_repo,
-                 grammar_processor):
+                 grammar_processor, snap_base_path, base, confinement,
+                 snap_type, soname_cache):
         self.valid = False
         self.plugin = plugin
-        self.config = {}
         self._part_properties = _expand_part_properties(
             part_properties, part_schema)
         self.stage_packages = []
         self._stage_packages_repo = stage_packages_repo
         self._grammar_processor = grammar_processor
+        self._snap_base_path = snap_base_path
+        self._base = base
+        self._confinement = confinement
+        self._snap_type = snap_type
+        self._soname_cache = soname_cache
+        self._source = grammar_processor.get_source()
+        if not self._source:
+            self._source = part_schema['source'].get('default')
+
+        self._pull_state = None  # type: states.PullState
+        self._build_state = None  # type: states.BuildState
+        self._stage_state = None  # type: states.StageState
+        self._prime_state = None  # type: states.PrimeState
 
         self._project_options = project_options
         self.deps = []
@@ -91,18 +99,56 @@ class PluginHandler:
         self._build_attributes = BuildAttributes(
             self._part_properties['build-attributes'])
 
+        # Scriptlet data is a dict of dicts for each step
+        self._scriptlet_metadata = collections.defaultdict(
+            snapcraft.extractors.ExtractedMetadata)
+        self._runner = Runner(
+            part_properties=self._part_properties,
+            sourcedir=self.plugin.sourcedir,
+            builddir=self.plugin.build_basedir,
+            stagedir=self.stagedir,
+            primedir=self.primedir,
+            builtin_functions={
+                'pull': self._do_pull,
+                'build': self.plugin.build,
+                'stage': self._do_stage,
+                'prime': self._do_prime,
+                'set-version': self._set_version,
+                'set-grade': self._set_grade,
+            })
+
         self._migrate_state_file()
 
+    def get_pull_state(self) -> states.PullState:
+        if not self._pull_state:
+            self._pull_state = states.get_state(self.plugin.statedir, 'pull')
+        return self._pull_state
+
+    def get_build_state(self) -> states.BuildState:
+        if not self._build_state:
+            self._build_state = states.get_state(self.plugin.statedir, 'build')
+        return self._build_state
+
+    def get_stage_state(self) -> states.StageState:
+        if not self._stage_state:
+            self._stage_state = states.get_state(self.plugin.statedir, 'stage')
+        return self._stage_state
+
+    def get_prime_state(self) -> states.PrimeState:
+        if not self._prime_state:
+            self._prime_state = states.get_state(self.plugin.statedir, 'prime')
+        return self._prime_state
+
     def _get_source_handler(self, properties):
         """Returns a source_handler for the source in properties."""
         # TODO: we cannot pop source as it is used by plugins. We also make
         # the default '.'
         source_handler = None
-        if properties['source']:
+        if self._source:
             handler_class = sources.get_source_handler(
-                properties['source'], source_type=properties['source-type'])
+                self._source, source_type=properties['source-type'])
             source_handler = handler_class(
-                properties['source'],
+                self._source,
                 self.plugin.sourcedir,
                 source_checksum=properties['source-checksum'],
                 source_branch=properties['source-branch'],
@@ -113,6 +159,46 @@ class PluginHandler:
 
         return source_handler
 
+    def _set_version(self, *, version):
+        try:
+            self._set_scriptlet_metadata(
+                snapcraft.extractors.ExtractedMetadata(version=version))
+        except errors.ScriptletDuplicateDataError as e:
+            raise errors.ScriptletDuplicateFieldError('version', e.other_step)
+
+    def _set_grade(self, *, grade):
+        try:
+            self._set_scriptlet_metadata(
+                snapcraft.extractors.ExtractedMetadata(grade=grade))
+        except errors.ScriptletDuplicateDataError as e:
+            raise errors.ScriptletDuplicateFieldError('grade', e.other_step)
+
+    def _set_scriptlet_metadata(
+            self, metadata: snapcraft.extractors.ExtractedMetadata):
+        step = self.next_step()
+
+        # First, ensure the metadata set here doesn't conflict with metadata
+        # already set for this step
+        conflicts = metadata.overlap(self._scriptlet_metadata[step])
+        if len(conflicts) > 0:
+            raise errors.ScriptletDuplicateDataError(
+                step, step, list(conflicts))
+
+        last_step = self.last_step()
+        if last_step:
+            # Now ensure the metadata from this step doesn't conflict with
+            # metadata from any other step
+            index = common.COMMAND_ORDER.index(last_step)
+            for index in reversed(range(0, index+1)):
+                other_step = common.COMMAND_ORDER[index]
+                state = states.get_state(self.plugin.statedir, other_step)
+                conflicts = metadata.overlap(state.scriptlet_metadata)
+                if len(conflicts) > 0:
+                    raise errors.ScriptletDuplicateDataError(
+                        step, other_step, list(conflicts))
+
+        self._scriptlet_metadata[step].update(metadata)
+
     def makedirs(self):
         dirs = [
             self.plugin.sourcedir, self.plugin.builddir,
@@ -146,6 +232,16 @@ class PluginHandler:
 
         return None
 
+    def next_step(self):
+        next_step = None
+        for step in reversed(common.COMMAND_ORDER):
+            if os.path.exists(
+                    states.get_step_state_file(self.plugin.statedir, step)):
+                break
+            next_step = step
+
+        return next_step
+
     def is_clean(self, step):
         """Return true if the given step hasn't run (or has been cleaned)."""
 
@@ -245,14 +341,25 @@ class PluginHandler:
         self._unpack_stage_packages()
 
     def pull(self, force=False):
+        # Ensure any previously-failed pull is cleared out before we try again
+        if (os.path.islink(self.plugin.sourcedir) or
+                os.path.isfile(self.plugin.sourcedir)):
+            os.remove(self.plugin.sourcedir)
+        elif os.path.isdir(self.plugin.sourcedir):
+            shutil.rmtree(self.plugin.sourcedir)
+
         self.makedirs()
         self.notify_part_progress('Pulling')
+
+        self._runner.pull()
+
+        self.mark_pull_done()
+
+    def _do_pull(self):
         if self.source_handler:
             self.source_handler.pull()
         self.plugin.pull()
 
-        self.mark_pull_done()
-
     def mark_pull_done(self):
         pull_properties = self.plugin.get_pull_properties()
 
@@ -260,13 +367,24 @@ class PluginHandler:
         part_build_packages = self._part_properties.get('build-packages', [])
         part_build_snaps = self._part_properties.get('build-snaps', [])
 
+        # Extract any requested metadata available in the source directory
+        metadata = snapcraft.extractors.ExtractedMetadata()
+        metadata_files = []
+        for path in self._part_properties.get('parse-info', []):
+            file_path = os.path.join(self.plugin.sourcedir, path)
+            with contextlib.suppress(errors.MissingMetadataFileError):
+                metadata.update(extract_metadata(self.name, file_path))
+                metadata_files.append(path)
+
         self.mark_done('pull', states.PullState(
             pull_properties, part_properties=self._part_properties,
             project=self._project_options, stage_packages=self.stage_packages,
             build_snaps=part_build_snaps,
             build_packages=part_build_packages,
-            source_details=self.source_handler.source_details
-        ))
+            source_details=self.source_handler.source_details,
+            metadata=metadata,
+            metadata_files=metadata_files,
+            scriptlet_metadata=self._scriptlet_metadata['pull']))
 
     def clean_pull(self, hint=''):
         if self.is_clean('pull'):
@@ -320,15 +438,19 @@ class PluginHandler:
         shutil.copytree(self.plugin.sourcedir, self.plugin.build_basedir,
                         symlinks=True, ignore=ignore)
 
-        script_runner = ScriptRunner(builddir=self.plugin.build_basedir)
+        self._runner.prepare()
+        self._runner.build()
+        self._runner.install()
 
-        script_runner.run(scriptlet=self._part_properties.get('prepare'))
-        build_scriptlet = self._part_properties.get('build')
-        if build_scriptlet:
-            script_runner.run(scriptlet=build_scriptlet)
-        else:
-            self.plugin.build()
-        script_runner.run(scriptlet=self._part_properties.get('install'))
+        # Organize the installed files as requested. We do this in the build
+        # step for two reasons:
+        #
+        #   1. So cleaning and re-running the stage step works even if
+        #      `organize` is used
+        #   2. So collision detection takes organization into account, i.e. we
+        #      can use organization to get around file collisions between
+        #      parts when staging.
+        self._organize()
 
         self.mark_build_done()
 
@@ -337,9 +459,41 @@ class PluginHandler:
         plugin_manifest = self.plugin.get_manifest()
         machine_manifest = self._get_machine_manifest()
 
+        # Extract any requested metadata available in the build directory,
+        # followed by the install directory (which takes precedence)
+        metadata_files = []
+        metadata = snapcraft.extractors.ExtractedMetadata()
+        for path in self._part_properties.get('parse-info', []):
+            file_path = os.path.join(self.plugin.builddir, path)
+            found = False
+            with contextlib.suppress(errors.MissingMetadataFileError):
+                metadata.update(extract_metadata(self.name, file_path))
+                found = True
+
+            file_path = os.path.join(self.plugin.installdir, path)
+            with contextlib.suppress(errors.MissingMetadataFileError):
+                metadata.update(extract_metadata(self.name, file_path))
+                found = True
+
+            if found:
+                metadata_files.append(path)
+            else:
+                # If this metadata file is not found in either build or
+                # install, check the pull state to make sure it was found
+                # there. If not, we need to let the user know.
+                state = self.get_pull_state()
+                if not state or path not in state.extracted_metadata['files']:
+                    raise errors.MissingMetadataFileError(self.name, path)
+
         self.mark_done('build', states.BuildState(
-            build_properties, self._part_properties, self._project_options,
-            plugin_manifest, machine_manifest))
+            property_names=build_properties,
+            part_properties=self._part_properties,
+            project=self._project_options,
+            plugin_assets=plugin_manifest,
+            machine_assets=machine_manifest,
+            metadata=metadata,
+            metadata_files=metadata_files,
+            scriptlet_metadata=self._scriptlet_metadata['build']))
 
     def _get_machine_manifest(self):
         return {
@@ -396,7 +550,14 @@ class PluginHandler:
     def stage(self, force=False):
         self.makedirs()
         self.notify_part_progress('Staging')
-        self._organize()
+        self._runner.stage()
+
+        # Only mark this step done if _do_stage() didn't run, in which case
+        # we have no directories or files to track.
+        if self.is_clean('stage'):
+            self.mark_stage_done(set(), set())
+
+    def _do_stage(self):
         snap_files, snap_dirs = self.migratable_fileset_for('stage')
 
         def fixup_func(file_path):
@@ -417,7 +578,7 @@ class PluginHandler:
     def mark_stage_done(self, snap_files, snap_dirs):
         self.mark_done('stage', states.StageState(
             snap_files, snap_dirs, self._part_properties,
-            self._project_options))
+            self._project_options, self._scriptlet_metadata['stage']))
 
     def clean_stage(self, project_staged_state, hint=''):
         if self.is_clean('stage'):
@@ -438,44 +599,71 @@ class PluginHandler:
 
         self.mark_cleaned('stage')
 
-    def prime(self, force=False):
+    def prime(self, force=False) -> None:
         self.makedirs()
         self.notify_part_progress('Priming')
+        self._runner.prime()
+
+        # Only mark this step done if _do_prime() didn't run, in which case
+        # we have no files, directories, or dependency paths to track.
+        if self.is_clean('prime'):
+            self.mark_prime_done(set(), set(), set())
+
+    def _do_prime(self) -> None:
         snap_files, snap_dirs = self.migratable_fileset_for('prime')
         _migrate_files(snap_files, snap_dirs, self.stagedir, self.primedir)
 
-        dependencies = _find_dependencies(self.primedir, snap_files)
-
-        # Split the necessary dependencies into their corresponding location.
-        # We'll both migrate and track the system dependencies, but we'll only
-        # track the part and staged dependencies, since they should have
-        # already been primed by other means, and migrating them again could
-        # potentially override the `stage` or `snap` filtering.
-        (in_part, staged, primed, system) = _split_dependencies(
-            dependencies, self.installdir, self.stagedir, self.primedir)
-
-        part_dependency_paths = {os.path.dirname(d) for d in in_part}
-        staged_dependency_paths = {os.path.dirname(d) for d in staged}
-
-        dependency_paths = part_dependency_paths | staged_dependency_paths
-
-        if not self._build_attributes.no_system_libraries():
-            system_dependency_paths = {os.path.dirname(d) for d in system}
-            dependency_paths.update(system_dependency_paths)
-
-            if system:
-                # Lots of dependencies are linked with a symlink, so we need to
-                # make sure we follow those symlinks when we migrate the
-                # dependencies.
-                _migrate_files(system, system_dependency_paths, '/',
-                               self.primedir, follow_symlinks=True)
+        if self._snap_type == 'app':
+            dependency_paths = self._handle_elf(snap_files)
+        else:
+            dependency_paths = set()
 
         self.mark_prime_done(snap_files, snap_dirs, dependency_paths)
 
+    def _handle_elf(self, snap_files: Sequence[str]) -> Set[str]:
+        elf_files = elf.get_elf_files(self.primedir, snap_files)
+        all_dependencies = set()
+        # TODO: base snap support
+        core_path = common.get_core_path(self._base)
+
+        # Clear the cache of all libs that aren't already in the primedir
+        self._soname_cache.reset_except_root(self.primedir)
+        for elf_file in elf_files:
+            all_dependencies.update(
+                elf_file.load_dependencies(root_path=self.primedir,
+                                           core_base_path=core_path,
+                                           soname_cache=self._soname_cache))
+
+        dependency_paths = self._handle_dependencies(all_dependencies)
+
+        if not self._build_attributes.keep_execstack():
+            clear_execstack(elf_files=elf_files)
+
+        if self._build_attributes.no_patchelf():
+            logger.warning(
+                'The primed files for part {!r} will not be verified for '
+                'correctness or patched: build-attributes: [no-patchelf] '
+                'is set.'.format(self.name))
+        else:
+            part_patcher = PartPatcher(
+                elf_files=elf_files,
+                plugin=self.plugin,
+                project=self._project_options,
+                confinement=self._confinement,
+                core_base=self._base,
+                snap_base_path=self._snap_base_path,
+                stagedir=self.stagedir,
+                primedir=self.primedir,
+                stage_packages=self._part_properties.get(
+                    'stage-packages', []))
+            part_patcher.patch()
+
+        return dependency_paths
+
     def mark_prime_done(self, snap_files, snap_dirs, dependency_paths):
         self.mark_done('prime', states.PrimeState(
             snap_files, snap_dirs, dependency_paths, self._part_properties,
-            self._project_options))
+            self._project_options, self._scriptlet_metadata['prime']))
 
     def clean_prime(self, project_primed_state, hint=''):
         if self.is_clean('prime'):
@@ -514,6 +702,39 @@ class PluginHandler:
         _clean_migrated_files(primed_files, primed_directories,
                               shared_directory)
 
+    def _handle_dependencies(self, all_dependencies: Set[str]):
+        # Split the necessary dependencies into their corresponding location.
+        # We'll both migrate and track the system dependencies, but we'll only
+        # track the part and staged dependencies, since they should have
+        # already been primed by other means, and migrating them again could
+        # potentially override the `stage` or `snap` filtering.
+        (in_part, staged, primed, system) = _split_dependencies(
+            all_dependencies, self.installdir, self.stagedir, self.primedir)
+        part_dependency_paths = {os.path.dirname(d) for d in in_part}
+        staged_dependency_paths = {os.path.dirname(d) for d in staged}
+        dependency_paths = part_dependency_paths | staged_dependency_paths
+
+        if not self._build_attributes.no_system_libraries():
+            system_dependency_paths = {os.path.dirname(d) for d in system}
+            dependency_paths.update(system_dependency_paths)
+
+            if system:
+                # Lots of dependencies are linked with a symlink, so we need to
+                # make sure we follow those symlinks when we migrate the
+                # dependencies.
+                _migrate_files(system, system_dependency_paths, '/',
+                               self.primedir, follow_symlinks=True)
+                formatted_system = '\n'.join(sorted(system))
+                logger.warning(
+                    'Files from the build host were migrated into the snap to '
+                    'satisfy dependencies that would otherwise not be met. '
+                    'This feature will be removed in a future release. If '
+                    'these libraries are needed in the final snap, ensure '
+                    'that the following are either satisfied by a '
+                    'stage-packages entry or through a part:\n{}'.format(
+                        formatted_system))
+        return dependency_paths
+
     def get_primed_dependency_paths(self):
         dependency_paths = set()
         state = states.get_state(self.plugin.statedir, 'prime')
@@ -710,7 +931,9 @@ def _migrate_files(snap_files, snap_dirs, srcdir, dstdir, missing_ok=False,
 def _organize_filesets(fileset, base_dir):
     for key in sorted(fileset, key=lambda x: ['*' in x, x]):
         src = os.path.join(base_dir, key)
-        dst = os.path.join(base_dir, fileset[key])
+        # Remove the leading slash if there so os.path.join
+        # actually joins
+        dst = os.path.join(base_dir, fileset[key].lstrip('/'))
 
         sources = iglob(src, recursive=True)
 
@@ -745,41 +968,6 @@ def _clean_migrated_files(snap_files, snap_dirs, directory):
             os.rmdir(migrated_directory)
 
 
-def _find_dependencies(root, part_files):
-    ms = magic.open(magic.NONE)
-    if ms.load() != 0:
-        raise RuntimeError('Cannot load magic header detection')
-
-    elf_files = set()
-
-    fs_encoding = sys.getfilesystemencoding()
-
-    for part_file in part_files:
-        # Filter out object (*.o) files-- we only care about binaries.
-        if part_file.endswith('.o'):
-            continue
-
-        # No need to crawl links-- the original should be here, too.
-        path = os.path.join(root, part_file)
-        if os.path.islink(path):
-            logger.debug('Skipped link {!r} while finding dependencies'.format(
-                path))
-            continue
-
-        path = path.encode(fs_encoding, errors='surrogateescape')
-        # Finally, make sure this is actually an ELF before queueing it up
-        # for an ldd call.
-        file_m = ms.file(path)
-        if file_m.startswith('ELF') and 'dynamically linked' in file_m:
-            elf_files.add(path)
-
-    dependencies = []
-    for elf_file in elf_files:
-        dependencies += libraries.get_dependencies(elf_file)
-
-    return set(dependencies)
-
-
 def _get_file_list(stage_set):
     includes = []
     excludes = []
