@@ -1,6 +1,6 @@
 # -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
 #
-# Copyright (C) 2016-2017 Canonical Ltd
+# Copyright (C) 2016-2018 Canonical Ltd
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License version 3 as
@@ -13,16 +13,13 @@
 #
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
-
 import logging
+from os import path
 
 import snapcraft
-from snapcraft.internal import (
-    deprecations,
-    pluginhandler,
-    repo
-)
+from snapcraft.internal import deprecations, elf, pluginhandler, repo
 from ._env import (
+    env_for_classic,
     build_env,
     build_env_for_stage,
     runtime_env,
@@ -40,8 +37,11 @@ class PartsConfig:
     def __init__(self, *, parts, project_options, validator,
                  build_snaps, build_tools, snapcraft_yaml):
         self._snap_name = parts['name']
-        self._confinement = parts['confinement']
+        self._base = parts.get('base', 'core')
+        self._confinement = parts.get('confinement')
+        self._soname_cache = elf.SonameCache()
         self._parts_data = parts.get('parts', {})
+        self._snap_type = parts.get('type', 'app')
         self._project_options = project_options
         self._validator = validator
         self.build_snaps = build_snaps
@@ -181,7 +181,7 @@ class PartsConfig:
         grammar_processor = grammar_processing.PartGrammarProcessor(
             plugin=plugin,
             properties=part_properties,
-            project_options=self._project_options,
+            project=self._project_options,
             repo=stage_packages_repo)
 
         part = pluginhandler.PluginHandler(
@@ -191,12 +191,21 @@ class PartsConfig:
             part_schema=self._validator.part_schema,
             definitions_schema=self._validator.definitions_schema,
             stage_packages_repo=stage_packages_repo,
-            grammar_processor=grammar_processor)
+            grammar_processor=grammar_processor,
+            snap_base_path=path.join('/', 'snap', self._snap_name, 'current'),
+            base=self._base,
+            confinement=self._confinement,
+            snap_type=self._snap_type,
+            soname_cache=self._soname_cache)
 
         self.build_snaps |= grammar_processor.get_build_snaps()
         self.build_tools |= grammar_processor.get_build_packages()
 
+        # TODO: this should not pass in command but the required package,
+        #       where the required package is to be determined by the
+        #       source handler.
         if part.source_handler and part.source_handler.command:
+            # TODO get_packages_for_source_type should not be a thing.
             self.build_tools |= repo.Repo.get_packages_for_source_type(
                 part.source_handler.command)
         self.all_parts.append(part)
@@ -208,7 +217,8 @@ class PartsConfig:
 
         env = []
         stagedir = self._project_options.stage_dir
-        core_dynamic_linker = self._project_options.get_core_dynamic_linker()
+        is_host_compat = self._project_options.is_host_compatible_with_base(
+            self._base)
 
         if root_part:
             # this has to come before any {}/usr/bin
@@ -220,16 +230,19 @@ class PartsConfig:
             env += build_env(
                 part.installdir,
                 self._snap_name,
-                self._confinement,
-                self._project_options.arch_triplet,
-                core_dynamic_linker=core_dynamic_linker)
+                self._project_options.arch_triplet)
             env += build_env_for_stage(
                 stagedir,
                 self._snap_name,
-                self._confinement,
-                self._project_options.arch_triplet,
-                core_dynamic_linker=core_dynamic_linker)
+                self._project_options.arch_triplet)
+            # Only set the paths to the base snap if we are building on the
+            # same host. Failing to do so will cause Segmentation Faults.
+            if (self._confinement == 'classic' and is_host_compat):
+                env += env_for_classic(self._base,
+                                       self._project_options.arch_triplet)
             env.append('SNAPCRAFT_PART_INSTALL="{}"'.format(part.installdir))
+            env.append('SNAPCRAFT_ARCH_TRIPLET="{}"'.format(
+                self._project_options.arch_triplet))
             env.append('SNAPCRAFT_PARALLEL_BUILD_COUNT={}'.format(
                        self._project_options.parallel_build_count))
         else:
