@@ -14,6 +14,7 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+from collections import OrderedDict
 import contextlib
 
 import snapcraft.internal.errors
@@ -57,10 +58,10 @@ class MissingSnapcraftYamlError(ProjectLoaderError):
 
 class YamlValidationError(ProjectLoaderError):
 
-    fmt = 'Issues while validating {snapcraft_yaml}: {message}'
+    fmt = 'Issues while validating {source}: {message}'
 
     @classmethod
-    def from_validation_error(cls, error):
+    def from_validation_error(cls, error, *, source='snapcraft.yaml'):
         """Take a jsonschema.ValidationError and create a SnapcraftSchemaError.
 
         The validation errors coming from jsonschema are a nightmare. This
@@ -84,10 +85,10 @@ class YamlValidationError(ProjectLoaderError):
         else:
             messages.append(error.message)
 
-        return cls(' '.join(messages))
+        return cls(' '.join(messages), source)
 
-    def __init__(self, message, snapcraft_yaml='snapcraft.yaml'):
-        super().__init__(message=message, snapcraft_yaml=snapcraft_yaml)
+    def __init__(self, message, source='snapcraft.yaml'):
+        super().__init__(message=message, source=source)
 
 
 class SnapcraftLogicError(ProjectLoaderError):
@@ -128,6 +129,27 @@ def _determine_cause(error):
         messages.append(
             error.schema['validation-failure'].format(key))
 
+    # anyOf failures might have usable context... try to improve them a bit
+    if error.validator == 'anyOf':
+        contextual_messages = OrderedDict()  # type: Dict[str, str]
+        for contextual_error in error.context:
+            key = contextual_error.schema_path.popleft()
+            if key not in contextual_messages:
+                contextual_messages[key] = []
+            message = contextual_error.message
+            if message:
+                # Sure it starts lower-case (not all messages do)
+                contextual_messages[key].append(
+                    message[0].lower() + message[1:])
+
+        oneOf_messages = []  # type: List[str]
+        for key, value in contextual_messages.items():
+            oneOf_messages.append(formatting_utils.humanize_list(
+                value, 'and', '{}'))
+
+        messages.append(formatting_utils.humanize_list(
+            oneOf_messages, 'or', '{}'))
+
     return ' '.join(messages)
 
 
