@@ -19,6 +19,7 @@ import glob
 import hashlib
 import logging
 import os
+import re
 import shutil
 import stat
 import string
@@ -26,13 +27,14 @@ import subprocess
 import sys
 import urllib
 import urllib.request
+from typing import Dict, Set, List, Tuple  # noqa: F401
 
 import apt
 from xml.etree import ElementTree
 
 import snapcraft
 from snapcraft import file_utils
-from snapcraft.internal import cache, repo, common, os_release
+from snapcraft.internal import cache, repo, common, os_release, sources
 from snapcraft.internal.indicators import is_dumb_terminal
 from ._base import BaseRepo
 from . import errors
@@ -52,7 +54,9 @@ deb http://${security}.ubuntu.com/${suffix} ${release}-security universe
 deb http://${security}.ubuntu.com/${suffix} ${release}-security multiverse
 '''
 _GEOIP_SERVER = "http://geoip.ubuntu.com/lookup"
-_library_list = dict()
+_library_list = dict()  # type: Dict[str, Set[str]]
+_HASHSUM_MISMATCH_PATTERN = re.compile(
+    r'(E:Failed to fetch.+Hash Sum mismatch)+')
 
 
 class _AptCache:
@@ -120,10 +124,28 @@ class _AptCache:
             logger.warning(
                 "Cannot find 'dpkg' command needed to support multiarch")
 
-        apt_cache = apt.Cache(rootdir=cache_dir, memonly=True)
-        apt_cache.update(fetch_progress=self.progress,
-                         sources_list=sources_list_file)
+        return self._create_cache(cache_dir, sources_list_file)
 
+    def _create_cache(self, cache_dir: str,
+                      sources_list_file: str) -> apt.Cache:
+        apt_cache = apt.Cache(rootdir=cache_dir, memonly=True)
+        try:
+            apt_cache.update(fetch_progress=self.progress,
+                             sources_list=sources_list_file)
+        except apt.cache.FetchFailedException as e:
+            # In case of a hashsum mismatch, clear the index and try again.
+            # Re-raise the error if that failed.
+            if re.match(_HASHSUM_MISMATCH_PATTERN, str(e)):
+                try:
+                    index = apt.apt_pkg.config.find_dir('Dir::State::Lists')
+                    shutil.rmtree(index)
+                    apt_cache = apt.Cache(rootdir=cache_dir, memonly=True)
+                    apt_cache.update(fetch_progress=self.progress,
+                                     sources_list=sources_list_file)
+                except apt.cache.FetchFailedException as retry:
+                    raise errors.CacheUpdateFailedError(str(retry))
+            else:
+                raise errors.CacheUpdateFailedError(str(e))
         return apt_cache
 
     @contextlib.contextmanager
@@ -153,6 +175,32 @@ class _AptCache:
 
         return _get_local_sources_list()
 
+    def fetch_binary(self, *, package_candidate, destination: str) -> None:
+        # This is a workaround for the overly verbose python-apt we use.
+        # There is an unreleased patch which once released could replace
+        # this code https://salsa.debian.org/apt-team/python-apt/commit/d122f9142df614dbb5f7644112280140dc155ecc  # noqa
+        # What follows is almost a tit for tat implementation of upstream's
+        # fetch_binary logic.
+        base = os.path.basename(package_candidate._records.filename)
+        destfile = os.path.join(destination, base)
+        if apt.package._file_is_same(destfile, package_candidate.size,
+                                     package_candidate._records.md5_hash):
+            logging.debug('Ignoring already existing file: {}'.format(
+                destfile))
+            return os.path.abspath(destfile)
+        acq = apt.apt_pkg.Acquire(self.progress)
+        acqfile = apt.apt_pkg.AcquireFile(
+            acq, package_candidate.uri, package_candidate._records.md5_hash,
+            package_candidate.size, base, destfile=destfile)
+        acq.run()
+
+        if acqfile.status != acqfile.STAT_DONE:
+            raise apt.package.FetchError(
+                "The item %r could not be fetched: %s" %
+                (acqfile.destfile, acqfile.error_text))
+
+        return os.path.abspath(destfile)
+
 
 class Ubuntu(BaseRepo):
 
@@ -163,7 +211,9 @@ class Ubuntu(BaseRepo):
             output = subprocess.check_output(
                 ['dpkg', '-L', package_name]).decode(
                     sys.getfilesystemencoding()).strip().split()
-            _library_list[package_name] = {i for i in output if 'lib' in i}
+            _library_list[package_name] = {
+                i for i in output
+                if ('lib' in i and os.path.isfile(i))}
 
         return _library_list[package_name].copy()
 
@@ -179,19 +229,31 @@ class Ubuntu(BaseRepo):
             packages = {'mercurial'}
         elif source_type == 'subversion' or source_type == 'svn':
             packages = {'subversion'}
+        elif source_type == 'rpm2cpio':
+            packages = {'rpm2cpio'}
+        elif source_type == '7zip':
+            packages = {'p7zip-full'}
         else:
             packages = set()
 
         return packages
 
     @classmethod
-    def install_build_packages(cls, package_names):
-        """Install build packages on the building machine.
+    def install_build_packages(cls, package_names: List[str]) -> List[str]:
+        """Install packages on the host required to build.
 
+        :param package_names: a list of package names to install.
+        :type package_names: a list of strings.
         :return: a list with the packages installed and their versions.
-
+        :rtype: list of strings.
+        :raises snapcraft.repo.errors.BuildPackageNotFoundError:
+            if one of the packages was not found.
+        :raises snapcraft.repo.errors.PackageBrokenError:
+            if dependencies for one of the packages cannot be resolved.
+        :raises snapcraft.repo.errors.BuildPackagesNotInstalledError:
+            if installing the packages on the host failed.
         """
-        new_packages = []
+        new_packages = []  # type: List[Tuple[str, str]]
         with apt.Cache() as apt_cache:
             try:
                 cls._mark_install(apt_cache, package_names)
@@ -207,7 +269,8 @@ class Ubuntu(BaseRepo):
                 for package in new_packages]
 
     @classmethod
-    def _mark_install(cls, apt_cache, package_names):
+    def _mark_install(cls, apt_cache: apt.Cache,
+                      package_names: List[str]) -> None:
         for name in package_names:
             if name.endswith(':any'):
                 name = name[:-4]
@@ -219,12 +282,27 @@ class Ubuntu(BaseRepo):
             try:
                 if version:
                     _set_pkg_version(apt_cache[name_arch], version)
-                apt_cache[name_arch].mark_install()
+                # Disable automatic resolving of broken packages here
+                # because if that fails it raises a SystemError and the
+                # API doesn't expose enough information about he problem.
+                # Instead we let apt-get show a verbose error message later.
+                apt_cache[name_arch].mark_install(auto_fix=False)
+                cls._verify_marked_install(apt_cache[name_arch])
             except KeyError:
                 raise errors.PackageNotFoundError(name)
 
     @classmethod
-    def _install_new_build_packages(cls, package_names):
+    def _verify_marked_install(cls, package: apt.Package):
+        if not package.installed and not package.marked_install:
+            broken_deps = []  # type: List[str]
+            for deps in package.candidate.dependencies:
+                for dep in deps:
+                    if not dep.target_versions:
+                        broken_deps.append(dep.name)
+            raise errors.PackageBrokenError(package.name, broken_deps)
+
+    @classmethod
+    def _install_new_build_packages(cls, package_names: List[str]) -> None:
         package_names.sort()
         logger.info(
             'Installing build dependencies: %s', ' '.join(package_names))
@@ -240,7 +318,10 @@ class Ubuntu(BaseRepo):
             apt_command.extend(['-o', 'Dpkg::Progress-Fancy=1'])
         apt_command.append('install')
 
-        subprocess.check_call(apt_command + package_names, env=env)
+        try:
+            subprocess.check_call(apt_command + package_names, env=env)
+        except subprocess.CalledProcessError:
+            raise errors.BuildPackagesNotInstalledError(packages=package_names)
 
         try:
             subprocess.check_call(['sudo', 'apt-mark', 'auto'] +
@@ -271,7 +352,7 @@ class Ubuntu(BaseRepo):
                             package.name, package.installed.version))
         return installed_packages
 
-    def __init__(self, rootdir, sources=None, project_options=None):
+    def __init__(self, rootdir, sources=None, project_options=None) -> None:
         super().__init__(rootdir)
         self._downloaddir = os.path.join(rootdir, 'download')
 
@@ -289,7 +370,7 @@ class Ubuntu(BaseRepo):
         with self._apt.archive(self._cache.base_dir) as apt_cache:
             return package_name in apt_cache
 
-    def get(self, package_names):
+    def get(self, package_names) -> None:
         with self._apt.archive(self._cache.base_dir) as apt_cache:
             self._mark_install(apt_cache, package_names)
             self._filter_base_packages(apt_cache, package_names)
@@ -344,8 +425,9 @@ class Ubuntu(BaseRepo):
         pkg_list = []
         for package in apt_cache.get_changes():
             pkg_list.append(str(package.candidate))
-            source = package.candidate.fetch_binary(
-                self._cache.packages_dir, progress=self._apt.progress)
+            source = self._apt.fetch_binary(
+                package_candidate=package.candidate,
+                destination=self._cache.packages_dir)
             destination = os.path.join(
                 self._downloaddir, os.path.basename(source))
             with contextlib.suppress(FileNotFoundError):
@@ -354,15 +436,11 @@ class Ubuntu(BaseRepo):
 
         return pkg_list
 
-    def unpack(self, unpackdir):
+    def unpack(self, unpackdir) -> None:
         pkgs_abs_path = glob.glob(os.path.join(self._downloaddir, '*.deb'))
         for pkg in pkgs_abs_path:
-            # TODO needs elegance and error control
-            try:
-                subprocess.check_call(
-                    ['dpkg-deb', '--extract', pkg, unpackdir])
-            except subprocess.CalledProcessError:
-                raise errors.UnpackError(pkg)
+            sources.Deb(None, None).provision(
+                unpackdir, src=pkg, clean_target=False, keep_deb=True)
         self.normalize(unpackdir)
 
     def _manifest_dep_names(self, apt_cache):
