@@ -18,6 +18,8 @@ from snapcraft.internal.os_release import OsRelease
 from ._platform import _is_deb_based
 from snapcraft.internal import errors
 
+from typing import List
+
 
 class RepoError(errors.SnapcraftError):
     pass
@@ -29,7 +31,27 @@ class NoNativeBackendError(RepoError):
            "You can however use 'snapcraft cleanbuild' with a container.")
 
     def __init__(self):
-        super().__init__(distro=OsRelease().name())
+        try:
+            distro = OsRelease().name()
+        except errors.OsReleaseNameError:
+            distro = 'this system'
+        super().__init__(distro=distro)
+
+
+class CacheUpdateFailedError(RepoError):
+
+    fmt = (
+        "Failed to update the package cache: "
+        "Some files could not be downloaded:{errors}"
+        "Check that the sources on your host are configured correctly."
+    )
+
+    def __init__(self, errors: str) -> None:
+        if errors:
+            errors = '\n\n{}\n\n'.format(errors.replace(', ', '\n'))
+        else:
+            errors = ' '
+        super().__init__(errors=errors)
 
 
 class BuildPackageNotFoundError(RepoError):
@@ -40,6 +62,22 @@ class BuildPackageNotFoundError(RepoError):
         super().__init__(package=package)
 
 
+class BuildPackagesNotInstalledError(RepoError):
+
+    fmt = "Could not install all requested build packages: {packages}"
+
+    def __init__(self, *, packages: List[str]) -> None:
+        super().__init__(packages=' '.join(packages))
+
+
+class PackageBrokenError(RepoError):
+
+    fmt = "The package {package} has unmet dependencies: {deps}"
+
+    def __init__(self, package: str, deps: List[str]) -> None:
+        super().__init__(package=package, deps=' '.join(deps))
+
+
 class PackageNotFoundError(RepoError):
 
     @property
@@ -71,6 +109,17 @@ class UnpackError(RepoError):
         super().__init__(package=package)
 
 
+class SnapUnavailableError(RepoError):
+
+    fmt = ('Failed to install or refresh a snap: {snap_name!r} does not exist '
+           'or is not available on the desired channel {snap_channel!r}. '
+           'Use `snap info {snap_name}` to get a list of channels the '
+           'snap is available on.')
+
+    def __init__(self, *, snap_name: str, snap_channel: str) -> None:
+        super().__init__(snap_name=snap_name, snap_channel=snap_channel)
+
+
 class SnapInstallError(RepoError):
 
     fmt = ('Error while installing snap {snap_name!r} from channel '
