@@ -35,6 +35,9 @@ Additionally, this plugin uses the following plugin-specific keywords:
       (list)
       A list of targets to `npm run`.
       These targets will be run in order, after `npm install`
+    - npm-flags:
+      (list)
+      A list of flags for npm.
     - node-package-manager
       (string; default: npm)
       The language package manager to use to drive installation
@@ -102,6 +105,15 @@ class NodePlugin(snapcraft.BasePlugin):
             },
             'default': []
         }
+        schema['properties']['npm-flags'] = {
+            'type': 'array',
+            'minitems': 1,
+            'uniqueItems': False,
+            'items': {
+                'type': 'string'
+            },
+            'default': []
+        }
 
         if 'required' in schema:
             del schema['required']
@@ -112,7 +124,7 @@ class NodePlugin(snapcraft.BasePlugin):
     def get_build_properties(cls):
         # Inform Snapcraft of the properties associated with building. If these
         # change in the YAML Snapcraft will consider the build step dirty.
-        return ['node-packages', 'npm-run']
+        return ['node-packages', 'npm-run', 'npm-flags']
 
     @classmethod
     def get_pull_properties(cls):
@@ -120,24 +132,35 @@ class NodePlugin(snapcraft.BasePlugin):
         # change in the YAML Snapcraft will consider the build step dirty.
         return ['node-engine', 'node-package-manager']
 
+    @property
+    def _nodejs_tar(self):
+        if self._nodejs_tar_handle is None:
+            self._nodejs_tar_handle = sources.Tar(self._nodejs_release_uri,
+                                                  self._npm_dir)
+        return self._nodejs_tar_handle
+
+    @property
+    def _yarn_tar(self):
+        if self._yarn_tar_handle is None:
+            self._yarn_tar_handle = sources.Tar(_YARN_URL, self._npm_dir)
+        return self._yarn_tar_handle
+
     def __init__(self, name, options, project):
         super().__init__(name, options, project)
         self._source_package_json = os.path.join(
             os.path.abspath(self.options.source), 'package.json')
         self._npm_dir = os.path.join(self.partdir, 'npm')
-        self._nodejs_tar = sources.Tar(get_nodejs_release(
-            self.options.node_engine, self.project.deb_arch), self._npm_dir)
-        if self.options.node_package_manager == 'yarn':
-            logger.warning(
-                'EXPERIMENTAL: use of yarn to manage packages is experimental')
-            self._yarn_tar = sources.Tar(_YARN_URL, self._npm_dir)
         self._manifest = collections.OrderedDict()
+        self._nodejs_release_uri = get_nodejs_release(
+            self.options.node_engine, self.project.deb_arch)
+        self._nodejs_tar_handle = None
+        self._yarn_tar_handle = None
 
     def pull(self):
         super().pull()
         os.makedirs(self._npm_dir, exist_ok=True)
         self._nodejs_tar.download()
-        if hasattr(self, '_yarn_tar'):
+        if self.options.node_package_manager == 'yarn':
             self._yarn_tar.download()
         # do the install in the pull phase to download all dependencies.
         if self.options.node_package_manager == 'npm':
@@ -175,14 +198,15 @@ class NodePlugin(snapcraft.BasePlugin):
     def _npm_install(self, rootdir):
         self._nodejs_tar.provision(
             self.installdir, clean_target=False, keep_tarball=True)
-        npm_install = ['npm', '--cache-min=Infinity', 'install']
+        npm_cmd = ['npm'] + self.options.npm_flags
+        npm_install = npm_cmd + ['--cache-min=Infinity', 'install']
         for pkg in self.options.node_packages:
             self.run(npm_install + ['--global'] + [pkg], cwd=rootdir)
         if os.path.exists(os.path.join(rootdir, 'package.json')):
             self.run(npm_install, cwd=rootdir)
             self.run(npm_install + ['--global'], cwd=rootdir)
         for target in self.options.npm_run:
-            self.run(['npm', 'run', target], cwd=rootdir)
+            self.run(npm_cmd + ['run', target], cwd=rootdir)
         return self._get_installed_node_packages('npm', self.installdir)
 
     def _yarn_install(self, rootdir):
@@ -191,6 +215,11 @@ class NodePlugin(snapcraft.BasePlugin):
         self._yarn_tar.provision(
             self._npm_dir, clean_target=False, keep_tarball=True)
         yarn_cmd = [os.path.join(self._npm_dir, 'bin', 'yarn')]
+        yarn_cmd.extend(self.options.npm_flags)
+        if 'http_proxy' in os.environ:
+            yarn_cmd.extend(['--proxy', os.environ['http_proxy']])
+        if 'https_proxy' in os.environ:
+            yarn_cmd.extend(['--https-proxy', os.environ['https_proxy']])
         flags = []
         if rootdir == self.builddir:
             yarn_add = yarn_cmd + ['global', 'add']
