@@ -1,6 +1,6 @@
 # -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
 #
-# Copyright (C) 2016-2017 Canonical Ltd
+# Copyright 2016-2017 Canonical Ltd
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License version 3 as
@@ -14,9 +14,19 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+import contextlib
+import logging
+from requests.packages import urllib3
 from simplejson.scanner import JSONDecodeError
+from typing import List  # noqa
 
 from snapcraft.internal.errors import SnapcraftError
+from snapcraft import formatting_utils
+
+
+logger = logging.getLogger(__name__)
+
+_STORE_STATUS_URL = 'https://status.snapcraft.io/'
 
 
 class StoreError(SnapcraftError):
@@ -26,6 +36,12 @@ class StoreError(SnapcraftError):
 
     """
 
+    def __init__(self, **kwargs):
+        with contextlib.suppress(KeyError, AttributeError):
+            logger.debug('Store error response: {}'.format(
+                kwargs['response'].__dict__))
+        super().__init__(**kwargs)
+
 
 class InvalidCredentialsError(StoreError):
 
@@ -38,15 +54,30 @@ class InvalidCredentialsError(StoreError):
 
 class LoginRequiredError(StoreError):
 
-    fmt = 'Cannot continue without logging in successfully.'
+    fmt = '{message}'
+
+    def __init__(self, extra_information=''):
+        message = 'Cannot continue without logging in successfully'
+        if extra_information:
+            message += ': {}'.format(extra_information)
+        super().__init__(message=message)
 
 
-class StoreRetryError(StoreError):
+class StoreNetworkError(StoreError):
 
-    fmt = 'There seems to be a network error: {error}'
+    fmt = 'There seems to be a network error: {message}'
 
     def __init__(self, exception):
-        super().__init__(error=str(exception))
+        message = str(exception)
+        with contextlib.suppress(IndexError):
+            underlying_exception = exception.args[0]
+            if isinstance(underlying_exception,
+                          urllib3.exceptions.MaxRetryError):
+                message = (
+                    'maximum retries exceeded trying to reach the store.\n'
+                    'Check your network connection, and check the store '
+                    'status at {}'.format(_STORE_STATUS_URL))
+        super().__init__(message=message)
 
 
 class SnapNotFoundError(StoreError):
@@ -84,10 +115,23 @@ class SHAMismatchError(StoreError):
 
 class StoreAuthenticationError(StoreError):
 
-    fmt = 'Authentication error: {}.'
+    fmt = 'Authentication error: {message}'
 
-    def __init__(self, message):
-        super().__init__(message=message)
+    def __init__(self, message, response=None):
+        # Unfortunately the store doesn't give us a consistent error response,
+        # so we'll check the ones of which we're aware.
+        with contextlib.suppress(AttributeError, JSONDecodeError):
+            response_json = response.json()
+            extra_error_message = ''
+            if 'error_message' in response_json:
+                extra_error_message = response_json['error_message']
+            elif 'message' in response_json:
+                extra_error_message = response_json['message']
+
+            if extra_error_message:
+                message += ': {}'.format(extra_error_message)
+
+        super().__init__(response=response, message=message)
 
 
 class StoreTwoFactorAuthenticationRequired(StoreAuthenticationError):
@@ -109,7 +153,8 @@ class DeveloperAgreementSignError(StoreError):
         'Text: {text!r}')
 
     def __init__(self, response):
-        super().__init__(reason=response.reason, text=response.text)
+        super().__init__(
+            response=response, reason=response.reason, text=response.text)
 
 
 class NeedTermsSignedError(StoreError):
@@ -139,7 +184,7 @@ class StoreAccountInformationError(StoreError):
                         'error_list'] if 'extra' in error]
         except JSONDecodeError:
             pass
-        super().__init__(error=error, extra=extra)
+        super().__init__(response=response, error=error, extra=extra)
 
 
 class StoreKeyRegistrationError(StoreError):
@@ -155,7 +200,7 @@ class StoreKeyRegistrationError(StoreError):
                     error['message'] for error in response_json['error_list'])
         except JSONDecodeError:
             pass
-        super().__init__(error=error)
+        super().__init__(response=response, error=error)
 
 
 class StoreRegistrationError(StoreError):
@@ -177,7 +222,8 @@ class StoreRegistrationError(StoreError):
     __FMT_RESERVED = (
         'The name {snap_name!r} is reserved.\n\n'
         'If you are the publisher most users expect for '
-        '{snap_name!r} then please claim the name at {register_name_url!r}')
+        '{snap_name!r} then please claim the name at {register_name_url!r}\n\n'
+        'Otherwise, please register another name.')
 
     __FMT_RETRY_WAIT = (
         'You must wait {retry_after} seconds before trying to register '
@@ -239,7 +285,8 @@ class StoreUploadError(StoreError):
         'Text: {text!r}')
 
     def __init__(self, response):
-        super().__init__(reason=response.reason, text=response.text)
+        super().__init__(
+            response=response, reason=response.reason, text=response.text)
 
 
 class StorePushError(StoreError):
@@ -265,8 +312,9 @@ class StorePushError(StoreError):
             except AttributeError:
                 response_json['text'] = 'error while pushing'
 
-        super().__init__(snap_name=snap_name, status_code=response.status_code,
-                         **response_json)
+        super().__init__(
+            response=response, snap_name=snap_name,
+            status_code=response.status_code, **response_json)
 
 
 class StoreReviewError(StoreError):
@@ -295,19 +343,24 @@ class StoreReviewError(StoreError):
 
     def __init__(self, result):
         self.fmt = self.__messages[result['code']]
+        additional = []
         errors = result.get('errors')
         if errors:
             for error in errors:
                 message = error.get('message')
                 if message:
-                    self.fmt = '{}\n  - {message}'.format(
-                        self.fmt, message=message)
+                    additional.append('  - {message}'.format(message=message))
+        if additional:
+            self.additional = '\n'.join(additional)
+            self.fmt += '\n{additional}'
         self.code = result['code']
         super().__init__()
 
 
 class StoreReleaseError(StoreError):
 
+    fmt = '{message}'
+
     __FMT_NOT_REGISTERED = (
         'Sorry, try `snapcraft register {snap_name}` before trying to '
         'release or choose an existing revision.')
@@ -316,7 +369,7 @@ class StoreReleaseError(StoreError):
         '{code}: {message}\n')
 
     __FMT_UNAUTHORIZED_OR_FORBIDDEN = (
-        'Received {status_code!r}: {text!r}')
+        'Received {status_code!r}: {text}')
 
     def __init__(self, snap_name, response):
         self.fmt_errors = {
@@ -324,14 +377,14 @@ class StoreReleaseError(StoreError):
             401: self.__fmt_error_401_or_403,
             403: self.__fmt_error_401_or_403,
             404: self.__fmt_error_404,
+            500: self.__fmt_error_500,
         }
 
         fmt_error = self.fmt_errors.get(
             response.status_code, self.__fmt_error_unknown)
 
-        self.fmt = fmt_error(response)
-
-        super().__init__(snap_name=snap_name)
+        super().__init__(response=response, message=fmt_error(response).format(
+            snap_name=snap_name))
 
     def __to_json(self, response):
         try:
@@ -358,6 +411,11 @@ class StoreReleaseError(StoreError):
         try:
             text = response.text
 
+            with contextlib.suppress(AttributeError, JSONDecodeError):
+                response_json = response.json()
+                if 'error_list' in response_json:
+                    text = _error_list_to_message(response_json)
+
         except AttributeError:
             text = 'error while releasing'
 
@@ -367,6 +425,17 @@ class StoreReleaseError(StoreError):
     def __fmt_error_404(self, response):
         return self.__FMT_NOT_REGISTERED
 
+    def __fmt_error_500(self, response):
+        response_json = self.__to_json(response)
+        message = ('The store encountered an internal error. The status of '
+                   'store and associated services can be checked at:\n'
+                   '{}'.format(_STORE_STATUS_URL))
+
+        if 'error_list' in response_json:
+            message = _error_list_to_message(response_json)
+
+        return message
+
     def __fmt_error_unknown(self, response):
         response_json = self.__to_json(response)
 
@@ -379,6 +448,49 @@ class StoreReleaseError(StoreError):
         return fmt
 
 
+class StoreMetadataError(StoreError):
+
+    __FMT_NOT_FOUND = (
+        "Sorry, updating the information on the store has failed, first run "
+        "`snapcraft register {snap_name}` and then "
+        "`snapcraft push <snap-file>`."
+    )
+
+    fmt = 'Received {status_code!r}: {text!r}'
+
+    def __init__(self, snap_name, response, metadata):
+        try:
+            response_json = response.json()
+        except (AttributeError, JSONDecodeError):
+            response_json = {}
+
+        if response.status_code == 404:
+            self.fmt = self.__FMT_NOT_FOUND
+        elif response.status_code == 409:
+            conflicts = [(error['extra']['name'], error)
+                         for error in response_json['error_list']
+                         if error['code'] == 'conflict']
+            parts = ["Metadata not pushed!"]
+            for field_name, error in sorted(conflicts):
+                sent = metadata.get(field_name)
+                parts.extend((
+                    "Conflict in {!r} field:".format(field_name),
+                    "    In snapcraft.yaml: {!r}".format(sent),
+                    "    In the Store:      {!r}".format(error['message']),
+                ))
+            parts.append(
+                "You can repeat the push-metadata command with "
+                "--force to force the local values into the Store")
+            self.parts = "\n".join(parts)
+            self.fmt = '{parts}'
+        elif 'error_list' in response_json:
+            response_json['text'] = response_json['error_list'][0]['message']
+
+        super().__init__(
+            response=response, snap_name=snap_name,
+            status_code=response.status_code, **response_json)
+
+
 class StoreValidationError(StoreError):
 
     fmt = 'Received error {status_code!r}: {text!r}'
@@ -392,7 +504,7 @@ class StoreValidationError(StoreError):
         except (AttributeError, JSONDecodeError):
             response_json = {'text': message or response}
 
-        super().__init__(status_code=response.status_code,
+        super().__init__(response=response, status_code=response.status_code,
                          **response_json)
 
 
@@ -410,7 +522,7 @@ class StoreSnapBuildError(StoreError):
         except JSONDecodeError:
             pass
 
-        super().__init__(error=error)
+        super().__init__(response=response, error=error)
 
 
 class StoreSnapRevisionsError(StoreError):
@@ -430,7 +542,7 @@ class StoreSnapRevisionsError(StoreError):
             pass
 
         super().__init__(
-            snap_id=snap_id, arch=arch or 'any arch',
+            response=response, snap_id=snap_id, arch=arch or 'any arch',
             series=series or 'any', error=error)
 
 
@@ -461,7 +573,7 @@ class StoreChannelClosingError(StoreError):
             error = '{} {}'.format(
                 response.status_code, response.reason)
 
-        super().__init__(error=error)
+        super().__init__(response=response, error=error)
 
 
 class StoreChannelClosingPermissionError(StoreError):
@@ -561,3 +673,52 @@ class SignBuildAssertionError(StoreError):
 
     def __init__(self, snap_name):
         super().__init__(snap_name=snap_name)
+
+
+class InvalidLoginConfig(StoreError):
+
+    fmt = 'Invalid login config: {error}'
+
+    def __init__(self, error):
+        super().__init__(error=error)
+
+
+def _error_list_to_message(response_json):
+    """Handle error list.
+
+    The error format is given here:
+    https://dashboard.snapcraft.io/docs/api/snap.html#format
+    """
+    messages = []
+    for error_list_item in response_json['error_list']:
+        messages.append(_error_list_item_to_message(
+            error_list_item, response_json))
+
+    return ', and '.join(messages)
+
+
+def _error_list_item_to_message(error_list_item, response_json):
+    """Handle error list item according to code.
+
+    The list of codes are here:
+    https://dashboard.snapcraft.io/docs/api/snap.html#codes
+    """
+    code = error_list_item['code']
+    message = ''
+    if code == 'macaroon-permission-required':
+        message = _handle_macaroon_permission_required(response_json)
+
+    if message:
+        return message
+    else:
+        return error_list_item['message']
+
+
+def _handle_macaroon_permission_required(response_json):
+    if 'permission' in response_json and 'channels' in response_json:
+        if response_json['permission'] == 'channel':
+            channels = response_json['channels']
+            return 'Lacking permission to release to channel(s) {}'.format(
+                formatting_utils.humanize_list(channels, 'and'))
+
+    return ''
